/*
 * Squashfs - a compressed read only filesystem for Linux
 *
 * Copyright (c) 2002, 2003, 2004, 2005, 2006, 2007, 2008
 * Phillip Lougher <phillip@lougher.demon.co.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * inode.c
 */

/*
 * This file implements code to create and read inodes from disk.
 *
 * Inodes in Squashfs are identified by a 48-bit inode which encodes the
 * location of the compressed metadata block containing the inode, and the byte
 * offset into that block where the inode is placed (<block, offset>).
 *
 * To maximise compression there are different inodes for each file type
 * (regular file, directory, device, etc.), the inode contents and length
 * varying with the type.
 *
 * To further maximise compression, two types of regular file inode and
 * directory inode are defined: inodes optimised for frequently occurring
 * regular files and directories, and extended types where extra
 * information has to be stored.
 */

#include <linux/fs.h>
#include <linux/vfs.h>
#include <linux/zlib.h>

#include "squashfs4_fs.h"
#include "squashfs4_fs_sb.h"
#include "squashfs4_fs_i.h"
#include "squashfs4.h"

/*
 * Initialise VFS inode with the base inode information common to all
 * Squashfs inode types.  Sqsh_ino contains the unswapped base inode
 * off disk.
 */
static int squashfs4_new_inode(struct super_block *sb, struct inode *inode,
				struct squashfs4_base_inode *sqsh_ino)
{
	int err;

	err = squashfs4_get_id(sb, le16_to_cpu(sqsh_ino->uid), &inode->i_uid);
	if (err)
		return err;

	err = squashfs4_get_id(sb, le16_to_cpu(sqsh_ino->guid), &inode->i_gid);
	if (err)
		return err;

	inode->i_ino = le32_to_cpu(sqsh_ino->inode_number);
	inode->i_mtime.tv_sec = le32_to_cpu(sqsh_ino->mtime);
	inode->i_atime.tv_sec = inode->i_mtime.tv_sec;
	inode->i_ctime.tv_sec = inode->i_mtime.tv_sec;
	inode->i_mode = le16_to_cpu(sqsh_ino->mode);
	inode->i_size = 0;

	return err;
}


struct inode *squashfs4_iget(struct super_block *sb, long long ino,
				unsigned int ino_number)
{
	struct inode *inode = iget_locked(sb, ino_number);
	int err;

	TRACE("Entered squashfs4_iget\n");

	if (!inode)
		return ERR_PTR(-ENOMEM);
	if (!(inode->i_state & I_NEW))
		return inode;

	err = squashfs4_read_inode(inode, ino);
	if (err) {
		iget_failed(inode);
		return ERR_PTR(err);
	}

	unlock_new_inode(inode);
	return inode;
}


/*
 * Initialise VFS inode by reading inode from inode table (compressed
 * metadata).  The format and amount of data read depends on type.
 */
int squashfs4_read_inode(struct inode *inode, long long ino)
{
	struct super_block *sb = inode->i_sb;
	struct squashfs4_sb_info *msblk = sb->s_fs_info;
	u64 block = SQUASHFS4_INODE_BLK(ino) + msblk->inode_table;
	int err, type, offset = SQUASHFS4_INODE_OFFSET(ino);
	union squashfs4_inode squashfs4_ino;
	struct squashfs4_base_inode *sqshb_ino = &squashfs4_ino.base;

	TRACE("Entered squashfs4_read_inode\n");

	/*
	 * Read inode base common to all inode types.
	 */
	err = squashfs4_read_metadata(sb, sqshb_ino, &block,
				&offset, sizeof(*sqshb_ino));
	if (err < 0)
		goto failed_read;

	err = squashfs4_new_inode(sb, inode, sqshb_ino);
	if (err)
		goto failed_read;

	block = SQUASHFS4_INODE_BLK(ino) + msblk->inode_table;
	offset = SQUASHFS4_INODE_OFFSET(ino);

	type = le16_to_cpu(sqshb_ino->inode_type);
	switch (type) {
	case SQUASHFS4_REG_TYPE: {
		unsigned int frag_offset, frag;
		int frag_size;
		u64 frag_blk;
		struct squashfs4_reg_inode *sqsh_ino = &squashfs4_ino.reg;

		err = squashfs4_read_metadata(sb, sqsh_ino, &block, &offset,
							sizeof(*sqsh_ino));
		if (err < 0)
			goto failed_read;

		frag = le32_to_cpu(sqsh_ino->fragment);
		if (frag != SQUASHFS4_INVALID_FRAG) {
			frag_offset = le32_to_cpu(sqsh_ino->offset);
			frag_size = squashfs4_frag_lookup(sb, frag, &frag_blk);
			if (frag_size < 0) {
				err = frag_size;
				goto failed_read;
			}
		} else {
			frag_blk = SQUASHFS4_INVALID_BLK;
			frag_size = 0;
			frag_offset = 0;
		}

		inode->i_nlink = 1;
		inode->i_size = le32_to_cpu(sqsh_ino->file_size);
		inode->i_fop = &generic_ro_fops;
		inode->i_mode |= S_IFREG;
		inode->i_blocks = ((inode->i_size - 1) >> 9) + 1;
		squashfs4_i(inode)->fragment_block = frag_blk;
		squashfs4_i(inode)->fragment_size = frag_size;
		squashfs4_i(inode)->fragment_offset = frag_offset;
		squashfs4_i(inode)->start = le32_to_cpu(sqsh_ino->start_block);
		squashfs4_i(inode)->block_list_start = block;
		squashfs4_i(inode)->offset = offset;
		inode->i_data.a_ops = &squashfs4_aops;

		TRACE("File inode %x:%x, start_block %llx, block_list_start "
			"%llx, offset %x\n", SQUASHFS4_INODE_BLK(ino),
			offset, squashfs4_i(inode)->start, block, offset);
		break;
	}
	case SQUASHFS4_LREG_TYPE: {
		unsigned int frag_offset, frag;
		int frag_size;
		u64 frag_blk;
		struct squashfs4_lreg_inode *sqsh_ino = &squashfs4_ino.lreg;

		err = squashfs4_read_metadata(sb, sqsh_ino, &block, &offset,
							sizeof(*sqsh_ino));
		if (err < 0)
			goto failed_read;

		frag = le32_to_cpu(sqsh_ino->fragment);
		if (frag != SQUASHFS4_INVALID_FRAG) {
			frag_offset = le32_to_cpu(sqsh_ino->offset);
			frag_size = squashfs4_frag_lookup(sb, frag, &frag_blk);
			if (frag_size < 0) {
				err = frag_size;
				goto failed_read;
			}
		} else {
			frag_blk = SQUASHFS4_INVALID_BLK;
			frag_size = 0;
			frag_offset = 0;
		}

		inode->i_nlink = le32_to_cpu(sqsh_ino->nlink);
		inode->i_size = le64_to_cpu(sqsh_ino->file_size);
		inode->i_fop = &generic_ro_fops;
		inode->i_mode |= S_IFREG;
		inode->i_blocks = ((inode->i_size -
				le64_to_cpu(sqsh_ino->sparse) - 1) >> 9) + 1;

		squashfs4_i(inode)->fragment_block = frag_blk;
		squashfs4_i(inode)->fragment_size = frag_size;
		squashfs4_i(inode)->fragment_offset = frag_offset;
		squashfs4_i(inode)->start = le64_to_cpu(sqsh_ino->start_block);
		squashfs4_i(inode)->block_list_start = block;
		squashfs4_i(inode)->offset = offset;
		inode->i_data.a_ops = &squashfs4_aops;

		TRACE("File inode %x:%x, start_block %llx, block_list_start "
			"%llx, offset %x\n", SQUASHFS4_INODE_BLK(ino),
			offset, squashfs4_i(inode)->start, block, offset);
		break;
	}
	case SQUASHFS4_DIR_TYPE: {
		struct squashfs4_dir_inode *sqsh_ino = &squashfs4_ino.dir;

		err = squashfs4_read_metadata(sb, sqsh_ino, &block, &offset,
				sizeof(*sqsh_ino));
		if (err < 0)
			goto failed_read;

		inode->i_nlink = le32_to_cpu(sqsh_ino->nlink);
		inode->i_size = le16_to_cpu(sqsh_ino->file_size);
		inode->i_op = &squashfs4_dir_inode_ops;
		inode->i_fop = &squashfs4_dir_ops;
		inode->i_mode |= S_IFDIR;
		squashfs4_i(inode)->start = le32_to_cpu(sqsh_ino->start_block);
		squashfs4_i(inode)->offset = le16_to_cpu(sqsh_ino->offset);
		squashfs4_i(inode)->dir_idx_cnt = 0;
		squashfs4_i(inode)->parent = le32_to_cpu(sqsh_ino->parent_inode);

		TRACE("Directory inode %x:%x, start_block %llx, offset %x\n",
				SQUASHFS4_INODE_BLK(ino), offset,
				squashfs4_i(inode)->start,
				le16_to_cpu(sqsh_ino->offset));
		break;
	}
	case SQUASHFS4_LDIR_TYPE: {
		struct squashfs4_ldir_inode *sqsh_ino = &squashfs4_ino.ldir;

		err = squashfs4_read_metadata(sb, sqsh_ino, &block, &offset,
				sizeof(*sqsh_ino));
		if (err < 0)
			goto failed_read;

		inode->i_nlink = le32_to_cpu(sqsh_ino->nlink);
		inode->i_size = le32_to_cpu(sqsh_ino->file_size);
		inode->i_op = &squashfs4_dir_inode_ops;
		inode->i_fop = &squashfs4_dir_ops;
		inode->i_mode |= S_IFDIR;
		squashfs4_i(inode)->start = le32_to_cpu(sqsh_ino->start_block);
		squashfs4_i(inode)->offset = le16_to_cpu(sqsh_ino->offset);
		squashfs4_i(inode)->dir_idx_start = block;
		squashfs4_i(inode)->dir_idx_offset = offset;
		squashfs4_i(inode)->dir_idx_cnt = le16_to_cpu(sqsh_ino->i_count);
		squashfs4_i(inode)->parent = le32_to_cpu(sqsh_ino->parent_inode);

		TRACE("Long directory inode %x:%x, start_block %llx, offset "
				"%x\n", SQUASHFS4_INODE_BLK(ino), offset,
				squashfs4_i(inode)->start,
				le16_to_cpu(sqsh_ino->offset));
		break;
	}
	case SQUASHFS4_SYMLINK_TYPE:
	case SQUASHFS4_LSYMLINK_TYPE: {
		struct squashfs4_symlink_inode *sqsh_ino = &squashfs4_ino.symlink;

		err = squashfs4_read_metadata(sb, sqsh_ino, &block, &offset,
				sizeof(*sqsh_ino));
		if (err < 0)
			goto failed_read;

		inode->i_nlink = le32_to_cpu(sqsh_ino->nlink);
		inode->i_size = le32_to_cpu(sqsh_ino->symlink_size);
		inode->i_op = &page_symlink_inode_operations;
		inode->i_data.a_ops = &squashfs4_symlink_aops;
		inode->i_mode |= S_IFLNK;
		squashfs4_i(inode)->start = block;
		squashfs4_i(inode)->offset = offset;

		TRACE("Symbolic link inode %x:%x, start_block %llx, offset "
				"%x\n", SQUASHFS4_INODE_BLK(ino), offset,
				block, offset);
		break;
	}
	case SQUASHFS4_BLKDEV_TYPE:
	case SQUASHFS4_CHRDEV_TYPE:
	case SQUASHFS4_LBLKDEV_TYPE:
	case SQUASHFS4_LCHRDEV_TYPE: {
		struct squashfs4_dev_inode *sqsh_ino = &squashfs4_ino.dev;
		unsigned int rdev;

		err = squashfs4_read_metadata(sb, sqsh_ino, &block, &offset,
				sizeof(*sqsh_ino));
		if (err < 0)
			goto failed_read;

		if (type == SQUASHFS4_CHRDEV_TYPE)
			inode->i_mode |= S_IFCHR;
		else
			inode->i_mode |= S_IFBLK;
		inode->i_nlink = le32_to_cpu(sqsh_ino->nlink);
		rdev = le32_to_cpu(sqsh_ino->rdev);
		init_special_inode(inode, inode->i_mode, new_decode_dev(rdev));

		TRACE("Device inode %x:%x, rdev %x\n",
				SQUASHFS4_INODE_BLK(ino), offset, rdev);
		break;
	}
	case SQUASHFS4_FIFO_TYPE:
	case SQUASHFS4_SOCKET_TYPE:
	case SQUASHFS4_LFIFO_TYPE:
	case SQUASHFS4_LSOCKET_TYPE: {
		struct squashfs4_ipc_inode *sqsh_ino = &squashfs4_ino.ipc;

		err = squashfs4_read_metadata(sb, sqsh_ino, &block, &offset,
				sizeof(*sqsh_ino));
		if (err < 0)
			goto failed_read;

		if (type == SQUASHFS4_FIFO_TYPE)
			inode->i_mode |= S_IFIFO;
		else
			inode->i_mode |= S_IFSOCK;
		inode->i_nlink = le32_to_cpu(sqsh_ino->nlink);
		init_special_inode(inode, inode->i_mode, 0);
		break;
	}
	default:
		ERROR("Unknown inode type %d in squashfs4_iget!\n", type);
		return -EINVAL;
	}

	return 0;

failed_read:
	ERROR("Unable to read inode 0x%llx\n", ino);
	return err;
}
